# -*- coding: utf-8 -*-
import json
from .tools import to_string_url
from .entity import Entity


class Edge(Entity):
    """
    This class describes an Edge object child of Aquarium class.
    """

    def set_data_variables(self, data={}):
        """
        Sets the data variables of the object

        :param      data:  The object edge from Aquarium API
        :type       data:  dictionary
        """
        super(Edge, self).set_data_variables(data=data)
        self._from = data.get('_from', '')
        self._to = data.get('_to', '')

    def create(self, type='', from_key='', to_key='', data={}):
        """
        Create an edge

        .. tip::
            The type of the edge is case sensitive ! By convention, all edges' type start with a capital letter

        :param      type:      The edge type
        :type       type:      string
        :param      from_key:  The source key
        :type       from_key:  string
        :param      to_key:    The destination key
        :type       to_key:    string
        :param      data:      The edge data
        :type       data:      dictionary, optional

        :returns:   Edge object
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        payload = dict(fromKey=from_key,
                       toKey=to_key,
                       type=type,
                       data=data)

        result = self.do_request('POST', 'edges', data=json.dumps(payload))
        result = self.parent.cast(result)
        return result

    def replace_data(self, data={}):
        """
        Replace the edge data with new ones

        .. danger::
            This is replacing all existing edge data. Meaning that you can loose data.


        :param      data:  The new edge data
        :type       data:  dictionary

        :returns:   Edge object
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        data = dict(data=data)
        result = self.do_request(
            'PUT', 'edges/'+self._key, data=json.dumps(data))
        result = self.parent.cast(result)
        return result

    def update_data(self, data={}):
        """
        Update the edge data by merging the existing ones with the new ones

        :param      data:  The new edge data
        :type       data:  dictionary

        :returns:   Edge object
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        data = dict(data=data)
        result = self.do_request(
            'PATCH', 'edges/'+self._key, data=json.dumps(data))
        result = self.parent.cast(result)
        return result

    def get(self, populate=False):
        """
        Get the edge by its _key

        :param      populate:  Populate `edge.createdBy` and `edge.updatedBy` with User object
        :type       populate:  boolean, optional

        :returns:   Edge object
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        result = self.do_request(
            'GET', 'edges/{0}/?populate={1}'.format(self._key, to_string_url(populate)))
        result = self.parent.cast(result)
        return result

    def delete(self):
        """
        Delete the edge

        .. danger::
            The edge will be completely deleted

        :returns:   Deleted edge object from API
        :rtype:     dictionary
        """
        result = self.do_request('DELETE', 'edges/' + self._key)
        return result



# -*- coding: utf-8 -*-
from .tools import pretty_print_format
from inspect import ismethod

class Element(object):
    def __init__(self, parent=None):
        self.parent=parent

    def __call__(self, data={}):
        inst=self.__class__()
        for key, value in data.items():
            if isinstance(value, dict):
                value=self.parent.cast(value)
            elif isinstance(value, list):
                value=[self.parent.cast(v) if isinstance(v, dict) else v for v in value]
            setattr(inst, key, value)
        return inst

    def __getattr__(self, name):
        attrs_names=list(vars(self).keys())
        msg=['"{0}" object has no attribute "{1}".'.format(self.__class__.__name__, name)]

        for n in attrs_names:
            instance = getattr(self, n)
            if hasattr(instance, name):
                if 'You should try to call this attribute under' not in msg:
                    msg.append('You should try to call this attribute under')
                if len(msg)>2:
                    msg.append('or')
                msg.append('.'.join([self.__class__.__name__, n, name]))
        raise AttributeError(' '.join(msg))

    def __str__(self):
        entity=vars(self).copy()
        entity.pop('parent', None)
        return '\n\n[%s]\n%s' % (self.__class__.__name__, pretty_print_format(entity, indent=4))

    def __repr__(self):
        return str(self)

    def pop(self, attr_name):
        value=getattr(self, attr_name)
        delattr(self, attr_name)
        return value



class Error(Exception):
    def __init__(self, response):
        url=response.url
        error=response.json().get('error')
        super(Error, self).__init__('{0} - url:{1}'.format(error, url))

#400
class RequestError(Error):
    pass
#401
class AuthentificationError(Error):
    pass
#403
class AutorisationError(Error):
    pass
#404
class PathNotFoundError(Error):
    pass
#409
class ConflictError(Error):
    pass
#413
class UploadExceedLimit(Error):
    pass
#500
class InternalError(Error):
    pass


# -*- coding: utf-8 -*-
import sys
import datetime
import pytz
import logging
logger = logging.getLogger(__name__)


class Utils():
    """
    This class is a utility class.
    """

    @staticmethod
    def duration(days = 0, hours = 0, minutes = 0):
        """
        Generate an ISO 8601 duration string.

        :param      days:     The number of days to add
        :type       days:     number
        :param      hours:    The number of hours to add
        :type       hours:    number
        :param      minutes:  The number of minutes to add
        :type       minutes:  number

        :returns:   ISO 8601 duration string
        :rtype:     string
        """
        return 'P{days}{time}{hours}{minutes}'.format(
            days='{days}D'.format(days=days) if days > 0 else '',
            time='T' if hours > 0 or minutes > 0 else '',
            hours='{hours}H'.format(hours=hours) if hours > 0 else '',
            minutes='{minutes}M'.format(minutes=minutes) if minutes > 0 else '')

    @staticmethod
    def date():
        """
        Generate an ISO 8601 date string.

        :returns:   ISO 8601 date string
        :rtype:     string
        """

        date = datetime.datetime.now(pytz.UTC)

        if sys.version_info[0] > 2:
            return date.isoformat(timespec='milliseconds')
        else:
            return date.isoformat()


# -*- coding: utf-8 -*-
from ..item import Item
from .. import DEFAULT_STATUSES


class Task(Item):
    """
    This class describes a Template object child of Item class.
    """

    def assign_to(self, user_key=''):
        """
        Assign the task to user

        :param      user_key:  The user key
        :type       user_key:  string

        :returns:   Created assigned edge object
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        result = self.parent.edge.create(
            type='Assigned', from_key=str(self._key), to_key=str(user_key))
        return result

    def unassign_from(self, user_key=''):
        """
        Unassign the task from user

        :param      user_key:  The user key
        :type       user_key:  string

        :returns:   Deleted assigned edge object
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        result = self.traverse(
            meshql='# -($Assigned)> 0,1 $User AND item._key == "{user_key}" VIEW edge'.format(user_key=user_key))

        if len(result) > 0:
            assigned_edge = self.parent.cast(result[0])
            assigned_edge.delete()
            return assigned_edge
        else:
            return None

    def add_timelog(self, user_key, comment='', date='', duration=''):
        """
        Add a timelog to the task

        :param      user_key:  The user key
        :type       user_key:  string
        :param      comment:   The comment
        :type       comment:   string
        :param      date:      The date of the timelog. Use :func:`~aquarium.utils.Utils.date` to generate an ISO date string.
        :type       date:      string ISO 8601 (date)
        :param      duration:  The duration of the timelog. Use :func:`~aquarium.utils.Utils.duration` to generate an ISO duration string.
        :type       duration:  string ISO 8601 (duration)

        :returns:   Dictionary of Item object and Edge object
        :rtype:     dictionary {item: :class:`~aquarium.item.Item`, edge: :class:`~aquarium.edge.Edge`}
        """

        if user_key == None:
            user_key = self.parent.me()._key

        data = dict(
            duration=duration,
            comment=comment,
            performedAt=date,
            performedBy=user_key
        )
        result = self.append(type='Job', data=data)
        return result

    def get_statuses(self):
        """
        Gets all the statuses for the task

        :returns:   The statuses
        :rtype:     dictionary
        """
        query = "# <($Child, 40)- path.vertices[*].type NONE == 'User' -($Child)> $Properties VIEW item.data.tasks_status"
        statuses = self.traverse(meshql=query)
        statuses_dct = dict()

        for status in statuses:
            if status:
                name = status.get('status')
                if name not in statuses_dct:
                    statuses_dct[name] = status
        result = statuses_dct or DEFAULT_STATUSES
        return result

    def get_subtasks(self, status='', name='', is_completed=True):
        """
        Gets the subtasks of the task

        :param      status:        The status used to filter
        :type       status:        boolean, optional
        :param      name:          The name used to filter
        :type       name:          boolean, optional
        :param      is_completed:  Get completed subtasks
        :type       is_completed:  boolean, optional

        :returns:   List of Task object and Edge object
        :rtype:     List of dictionary {item: :class:`~aquarium.items.task.Task`, edge: :class:`~aquarium.edge.Edge`}
        """
        query = list()
        query.append("# -($Child)> ($Task")
        query.append(" AND item.data.completion {0}= 1".format(
            ['!', '='][is_completed]))
        if status:
            query.append("AND item.data.status == '{0}'".format(status))
        if name:
            query.append("AND item.data.name == '{0}'".format(name))

        query.append(')')

        result = self.traverse(meshql=' '.join(query))
        result = [self.parent.element(data) for data in result]
        return result

    def get_dependencies(self, mode='BOTH'):
        """
        Gets the dependencies of the task

        :param      mode:  The mode ("BOTH", "IN" or "OUT"). Used to get incoming dependencies, outgoing or both.
        :type       mode:  string, optional

        :returns:   List of Item object and Edge object
        :rtype:     List of dictionary {item: :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`, edge: :class:`~aquarium.edge.Edge}`
        """
        if mode == 'BOTH':
            query = "# <($Dependency)> *"
        elif mode == 'OUT':
            query = "# -($Dependency)> *"
        elif mode == 'IN':
            query = "# <($Dependency)- *"
        else:
            raise RuntimeError(
                'Wrong value for "mode". Use "BOTH", "IN" or "OUT"')

        result = self.traverse(meshql=query)
        result = [self.parent.element(data) for data in result]
        return result

    def get_assigned_users(self):
        """
        Gets all the assigned users to the task

        :returns:   List of User or Usergroup object and Edge object
        :rtype:     List of dictionary {item: :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`, edge: :class:`~aquarium.edge.Edge}`
        """
        query = "# -($Assigned)> *"
        result = self.traverse(meshql=query)
        result = [self.parent.element(data) for data in result]
        return result

    def get_attachments(self):
        """
        Gets all the task's attachments

        :returns:   List of Item object and Edge object
        :rtype:     List of dictionary {item: :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`, edge: :class:`~aquarium.edge.Edge`}
        """
        query = "# -($Attached)> *"
        result = self.traverse(meshql=query)
        result = [self.parent.element(data) for data in result]
        return result



# -*- coding: utf-8 -*-
from .asset import Asset


class Shot(Asset):
    """
    This class describes a Shot object child of Item class.

    .. note::
        This class is using all the same methods than the :class:`~aquarium.items.asset.Asset` class
    """
    pass



# -*- coding: utf-8 -*-
import json
from ..item import Item


class Organisation(Item):
    """
    This class describes an Organisation object child of Item class.
    """

    def get_member_by_email(self, email):
        """
        Get an exising member of the organisation by his/her email

        :returns:   User object
        :rtype:     :class:`~aquarium.items.user.User`
        """
        member = None
        members = self.get_all_members()
        filtered = [member for member in members if member.data.email == email]

        if (len(filtered) > 0):
            member = filtered[0]

        return member

    def get_all_members(self):
        """
        Gets all members of the organisation

        :returns:   List of User object
        :rtype:     List of :class:`~aquarium.items.user.User`
        """

        result=self.do_request('GET', 'organisations/{0}/members/all'.format(self._key))
        result=[self.parent.user(user) for user in result]
        return result

    def get_active_members(self):
        """
        Gets all active members of the organisation

        :returns:   List of User object
        :rtype:     List of :class:`~aquarium.items.user.User`
        """

        result=self.do_request('GET', 'organisations/{0}/members/active'.format(self._key))
        result=[self.parent.user(user) for user in result]
        return result

    def get_inactive_members(self):
        """
        Gets all inactive members of the organisation

        :returns:   List of User object
        :rtype:     List of :class:`~aquarium.items.user.User`
        """

        result=self.do_request('GET', 'organisations/{0}/members/inactive'.format(self._key))
        result=[self.parent.user(user) for user in result]
        return result

    def create_member(self, email, name=None):
        """
        Create a new member in your organisation

        :param      email:  The email of the new member
        :type       type:  string
        :param      name:  The name of the new member
        :type       type:  string, optional

        :returns:   User object
        :rtype:     :class:`~aquarium.items.user.User`
        """

        payload = dict(email=email, name=name)
        data = dict(data=payload)
        member = self.do_request(
            'POST', 'organisations/{0}/createMember'.format(self._key), data=json.dumps(data))

        member = self.parent.cast(member)
        return member



# -*- coding: utf-8 -*-
from . import URL_CONTENT_TYPE
import json
import re
import os
from .tools import to_string_url
from .entity import Entity
import logging
logger = logging.getLogger(__name__)


class Item(Entity):
    """
    This class describes an Item object child of Aquarium class.
    """

    def create(self, type='', data={}, path=None):
        """
        Create an item

        .. warning::
            We advice you to use :func:`~aquarium.item.Item.append` instead of :func:`~aquarium.item.Item.create`.
            This function will create an item, without inserting it to an existing hiearchy.
            Once the item is created, you should use Edge :func:`~aquarium.edge.Edge.create` function to connect it to an other existing item.
        .. tip::
            The type of the item is case sensitive ! By convention, all items' type start with a capital letter

        :param      type:  The newitem type
        :type       type:  string
        :param      data:  The newitem data
        :type       data:  dictionary, optional
        :param      path:  File path you want to upload on the appended item
        :type       path:  string, optional

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`
        """

        payload = dict(type=type, data=data)
        result = self.do_request('POST', 'items', data=json.dumps(payload))
        result = self.parent.cast(result)

        if path != None:
            upload = result.upload_file(path=path)
            if upload != None:
                for key in upload.data:
                    result.data[key] = upload.data[key]
        return result

    def append(self, type='', data={}, edge_type='Child', edge_data={}, apply_template=None, template_key=None, path=None):
        """
        Create and append a new item to the current one

        .. tip::
            The type of the item is case sensitive ! By convention, all items' type start with a capital letter


        :param      type:            The new item type
        :type       type:            string
        :param      data:            The new item data, optional
        :type       data:            dictionary
        :param      apply_template:  Do apply template ?
        :type       apply_template:  boolean, optional
        :param      template_key:    The template key to apply. If no template key, `automatic context's template <https://docs.fatfish.app/#/userguide/items?id=templates>`_ will be used.
        :type       template_key:    string, optional
        :param      path:            File path you want to upload on the appended item
        :type       path:            string, optional

        :returns:   Dictionary composed by an item and its edge.
        :rtype:     dict {item: :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup` | :class:`~aquarium.items.organisation.Organisation`, edge: :class:`~aquarium.edge.Edge}`
        """

        payload = {
            "item": {
                "type": type,
                "data": data
            },
            "edge": {
                "type": edge_type,
                "data": edge_data
            }
        }

        if apply_template is not None:
            payload["applyTemplate"] = apply_template

        if template_key:
            payload["templateKey"] = template_key

        result = self.do_request(
            'POST', 'items/'+self._key+'/append', data=json.dumps(payload))
        result = self.parent.element(result)

        if path != None:
            upload = result.item.upload_file(path=path)
            if upload != None:
                for key in upload.data:
                    result.item.data[key] = upload.data[key]

        return result

    def traverse(self, meshql='', aliases={}):
        """
        Execute a traverse from the current item

        :param      meshql:        The meshql string
        :type       meshql:        string
        :param      aliases:       The aliases used in the meshql query
        :type       aliases:       dictionary, optional

        :returns:   List of item and/or edge or VIEW used in the meshql query
        :rtype:     list
        """
        logger.debug('Send traverse : meshql : %s / aliases : %r',
                     meshql, aliases)
        data = dict(query=meshql, aliases=aliases)
        result = self.do_request(
            'POST', 'items/'+self._key+'/traverse', data=json.dumps(data))
        return result

    def replace_data(self, data={}):
        """
        Replace the item data with new ones

        .. danger::
            This is replacing all existing item data. Meaning that you can loose data.


        :param      data:  The new item data
        :type       data:  dictionary

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item`
        """
        logger.debug('Replacing data on item %s with %r', self._key, data)
        data = dict(data=data)
        result = self.do_request(
            'PUT', 'items/'+self._key, data=json.dumps(data))

        result = self.parent.cast(result)
        return result

    def update_data(self, data={}):
        """
        Update the item data by merging the existing ones with the new ones

        :param      data:  The new item data
        :type       data:  dictionary

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item`
        """
        logger.debug('Updating data on item %s with %r', self._key, data)
        data = dict(data=data)
        result = self.do_request(
            'PATCH', 'items/'+self._key, data=json.dumps(data))
        result = self.parent.cast(result)
        return result

    def copy(self, parent_key=''):
        """
        Copie the item into the parent

        :param      parent_key:  The key of the parent
        :type       parent_key:  string

        :returns:   List of new items object
        :rtype:     List of :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`
        """
        logger.debug('Copy item %s into %s', self._key, parent_key)
        data = dict(targetKey=parent_key)
        result = self.do_request(
            'POST', 'items/'+self._key+'/copy', data=json.dumps(data))

        result = [self.parent.cast(data) for data in result]
        return result

    def convert_to_template(self, parent_key=''):
        """
        Convert an item and its hierarchy to a template into a parent

        :param      parent_key:  The key of the parent
        :type       parent_key:  string

        :returns:   template object
        :rtype:     :class:`~aquarium.items.template.Template`
        """
        logger.debug('Converting item %s to template in %s',
                     self._key, parent_key)
        data = dict(parentKey=parent_key)
        result = self.do_request(
            'POST', 'items/'+self._key+'/convertToTemplate', data=json.dumps(data))
        result = self.parent.cast(result)
        return result

    def apply_template(self, template_key=''):
        """
        Apply a template on the item

        :param      template_key:  The key of the template
        :type       template_key:  string

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`
        """
        logger.debug('Apply template %s on item %s', template_key, self._key)
        data = dict(templateKey=template_key)
        result = self.do_request(
            'POST', 'items/'+self._key+'/template', data=json.dumps(data))
        result = self.parent.cast(result)
        return result

    def reapply_template(self, template_key=''):
        """
        Re-apply the template previously used

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`
        """
        logger.debug('Re-apply existing template on item %s', self._key)
        result = self.do_request(
            'POST', 'items/' + self._key + '/template/reapply')
        result = self.parent.cast(result)
        return result

    def get(self, populate=False, versions=False):
        """
        Get item object with its _key

        :param      populate:  Populate `item.createdBy` and `item.updatedBy` with User object
        :type       populate:  boolean
        :param      versions:  Get previous item's data versions
        :type       versions:  boolean, optional

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`
        """
        result = self.do_request('GET', 'items/{0}/?populate={1}&versions={2}'.format(
            self._key, int(populate), int(versions)), headers=URL_CONTENT_TYPE)
        result = self.parent.cast(result)
        return result

    def get_versions(self, populate=False):
        """
        Get the previous item's data versions

        :param      populate:  Populate `item.createdBy` and `item.updatedBy` with User object
        :type       populate:  boolean, optional

        :returns:   The versions values
        :rtype:     list
        """
        result = self.do_request(
            'GET', 'items/{0}/versions?populate={1}'.format(self._key, to_string_url(populate)))
        result = [self.parent.cast(data) for data in result]
        return result

    def get_shortest_path(self, key=''):
        """
        Get the shortest path between the current item and the item _key

        :param      key:  The destination key
        :type       key:  string

        :returns:   List of item Object
        :rtype:     list of :class:`~aquarium.item.Item` or subclass : :class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`
        """
        result = self.do_request(
            'GET', 'items/'+self._key+'/path/'+key, headers=URL_CONTENT_TYPE)
        result = [self.parent.cast(data) for data in result]
        return result

    def get_permissions(self, filters={}, populate=False):
        """
        Gets the permissions of the item

        :param      filters:   The filters
        :type       filters:   dictionary, optional
        :param      populate:  Populate with User object
        :type       populate:  boolean, optional

        :returns:   List of edge and user
        :rtype:     list
        """
        result = self.do_request('GET', 'items/{0}/permissions?filters={1}&populate={2}'.format(
            self._key, str(filters).replace("'", '"'), to_string_url(populate)), headers=URL_CONTENT_TYPE)
        result = [self.parent.element(data) for data in result]
        return result

    def get_parents(self):
        """
        Gets the parents of the item

        :returns:   List of item and edge object
        :rtype:     list of {item: :class:`~aquarium.item.Item`, edge: :class:`~aquarium.edge.Edge`}
        """
        query = "# <($Child)- *"
        result = self.traverse(meshql=query)
        result = [self.parent.element(data) for data in result]
        return result

    def get_children(self, show_hidden=False, types=None, names=None):
        """
        Gets the children of the item

        :param      show_hidden:  Show hidden items
        :type       show_hidden:  boolean, optional
        :param      types:  One string or list of string items type you want to filter
        :type       types:  string or list, optional
        :param      names:  One string or list of string items name you want to filter
        :type       names:  string or list, optional

        :returns:   List of item and edge object
        :rtype:     list of {item: :class:`~aquarium.item.Item`, edge: :class:`~aquarium.edge.Edge`}
        """
        query = ["# -($Child)>"]
        aliases = dict()

        if types == None:
            query.append('*')
        else:
            query.append('item.type IN @types')
            if not isinstance(types, list): types=[types]
            aliases['types'] = types

        if names != None:
            query.append('AND item.data.name IN @names')
            if not isinstance(names, list): names=[names]
            aliases['names'] = names

        if not show_hidden:
            query.append(' AND edge.data.isHidden != true')

        result = self.traverse(meshql=' '.join(query), aliases=aliases)
        result = [self.parent.element(data) for data in result]
        return result

    def get_trash(self):
        """
        Gets the trashed items

        :returns:   List of trashed item and edge object
        :rtype:     list of {item: :class:`~aquarium.item.Item`, edge: :class:`~aquarium.edge.Edge`}
        """
        query = '# -($Trash)> *'
        result = self.traverse(meshql=query)
        result = [self.parent.cast(data['item']) for data in result]
        return result

    def move(self, old_parent_key=None, new_parent_key=None):
        """
        Move item from old parent to new parent

        :param      old_parent_key:  The key of the old parent
        :type       old_parent_key:  string
        :param      new_parent_key:  The key of the new parent
        :type       new_parent_key:  string

        :returns:   New item parent and new child edge
        :rtype:     dictionary {item: :class:`~aquarium.item.Item`,edge: :class:`~aquarium.edge.Edge`}
        """
        logger.debug('Move item %s from parent %s to %s',
                     self._key, old_parent_key, new_parent_key)
        data = dict(
            oldParentKey=old_parent_key,
            newParentKey=new_parent_key
        )

        result = self.do_request(
            'PUT', 'items/'+self._key+'/move', data=json.dumps(data))
        result = self.parent.element(result)
        return result

    def trash(self, parent_key=''):
        """
        Move item from parent item to trash

        :param      parent_key:  The key of the parent
        :type       parent_key:  string

        :returns:   Trashed item
        :rtype:     dictionary
        """
        logger.debug('Trash item %s from parent %s', self._key, parent_key)
        data = dict(itemKey=self._key)
        result = self.do_request(
            'POST', 'items/'+parent_key+'/trash', data=json.dumps(data))
        result = self.parent.element(result)
        return result

    def restore(self, parent_key=''):
        """
        Restore an item from trash to parent item

        :param      parent_key:  The key of the parent
        :type       parent_key:  string

        :returns:   Restored edge
        :rtype:     :class:`~aquarium.edge.Edge`
        """
        logger.debug('Restore item %s from parent %s', self._key, parent_key)
        data = dict(itemKey=self._key)
        result = self.do_request(
            'POST', 'items/'+parent_key+'/restore', data=json.dumps(data))
        result = self.parent.cast(result)
        return result

    def delete(self):
        """
        Delete the item.

        .. danger::
            The item will be completely deleted. You can use :func:`~aquarium.item.Item.trash` instead

        :returns:   Deleted item object from API
        :rtype:     dictionary
        """
        logger.debug('Delete item %s', self._key)
        result = self.do_request(
            'DELETE', 'items/'+self._key, headers=URL_CONTENT_TYPE)
        return result

    def upload_file(self, path='', data = {}, message = None):
        """
        Upload a file on the item

        .. warning::
            This function will replace the data on the item.
            It's here to replace the existing file's data by creating a new history entry.
            We advice you to use :func:`~aquarium.item.Item.append` if you want to upload the file as a new item.

        :param      path:  The path of the file to upload
        :type       path:  string
        :param      data:  The data you want to upload with the file, optional
        :type       data:  dict
        :param      message:  The message associated with the upload, optional
        :type       message:  string

        :returns:   Item object
        :rtype:     :class:`~aquarium.item.Item`
        """
        logger.debug('Upload file %s on item %s with data %s', path, self._key, data)

        files = dict(
            file=open(path, 'rb'),
            data=(None, json.dumps(data), 'text/plain'),
            message=(None, message, 'text/plain')
        )
        result = self.do_request(
            'POST', 'items/'+self._key+'/upload', headers={'Content-Type': None}, files=files)
        files['file'].close()
        result = self.parent.cast(result)
        return result

    def download_file(self, path, versionKey=None):
        """
        Download the item's file to the path

        :param      path:  The path used to store the download. If directory is provided, the file name from original file is used
        :type       path:  string
        :param      versionKey:  The versionKey used to download the file
        :type       versionKey:  string, optional
        """
        logger.debug('Download from item %s to file %s', self._key, path)

        url = 'items/{_key}/download'.format(_key=self._key)
        if (versionKey != None):
            url = '{url}?versionKey=${versionKey}'.format(versionKey=versionKey)

        result = self.do_request(
            'GET', url, headers=URL_CONTENT_TYPE, decoding=False)

        if (os.path.isdir(path)):
            content_disposition = result.headers['content-disposition']
            filename = re.findall('filename="(.+)"', content_disposition)
            print(content_disposition)
            if len(filename) > 0:
                path = os.path.join(path, str(filename[0]))

        with open(path, 'wb') as f:
            for chunk in result:
                f.write(chunk)

        return path

    def import_json(self, content={}):
        """
        Import item hierarchy from json content

        :param      content:  The json content
        :type       content:  dictionary

        :returns:   Dictionary of imported items and edges
        :rtype:     dictionary
        """
        result = self.do_request(
            'POST', 'items/'+self._key+'/import/json', data=json.dumps(content))
        return result

    def export_json(self):
        """
        Export item hierarchy to json

        :returns:   Exported items and edges
        :rtype:     dictionary
        """
        result = self.do_request('GET', 'items/'+self._key+'/export/json')
        return result

    def compare(self, key=''):
        """
        Compare the item with a given item

        :param      key:  The key of the item to compare with
        :type       key:  string

        :returns:   The comparison result
        :rtype:     dictionary
        """
        result = self.do_request('POST', 'items/'+self._key+'/compare/'+key)
        result = self.parent.element(result)
        return result



# -*- coding: utf-8 -*-
from .. import URL_CONTENT_TYPE
from ..item import Item
from ..element import Element
import logging
logger = logging.getLogger(__name__)


class User(Item):
    """
    This class describes an User object child of Item class.
    """

    def signin(self, email='', password=''):
        """
        Sign in a user with it's email and password

        :param      email:     The email of the user
        :type       email:     string
        :param      password:  The password of the user
        :type       password:  string

        :returns: Dictionary User object
        :rtype: Dict {user: :class:`~aquarium.items.user.User`}
        """
        logging.debug('Connect user %s', email)
        # Authenticate and retrieve the access token
        payload = dict(email=email, password=password)
        result = self.do_request(
            'POST', 'signin', headers=URL_CONTENT_TYPE, data=payload)

       # Store authentification information
        token = result.pop("token")
        self.parent.token = token
        result = self.parent.element(result)

        return result

    def connect(self, email='', password=''):
        """
        Alias of :func:`~aquarium.items.user.User.signin`
        """
        return self.signin(email, password)

    def signout(self):
        """
        Sign out the current user by clearing the stored authentication token

        .. note::
            After a :func:`~aquarium.items.user.User.signout`, you need to use a :func:`~aquarium.items.user.User.signin` before sending authenticated requests

        :returns: None
        """
        self.do_request(
            'POST', 'signout', decoding=False)

       # Remove authentification information
        self.parent.token = None

        return None

    def get_profile(self):
        """
        Get the current profil.

        :returns:   User, Usergroups and Organisations object
        :rtype:     Dict {user: :class:`~aquarium.items.user.User`, usergroups: [:class:`~aquarium.items.usergroup.Usergroup`], organisations: [:class:`~aquarium.items.organisation.Organisation`]}
        """
        result = self.do_request('GET', 'users/me')
        result = self.parent.element(result)
        return result

    def get_current(self):
        """
        Get the current user.

        :returns:   User object
        :rtype:     :class:`~aquarium.items.user.User`
        """
        result = self.get_profile().user
        return result

    def get_tasks(self, project_key='', task_status='', task_name='', task_completed=False):
        """
        Gets the assigned task of the user

        :param      project_key:      The project key used to filter
        :type       project_key:      string
        :param      task_status:      The task status used to filter
        :type       task_status:      string, optional
        :param      task_name:        The task nam used to filtere
        :type       task_name:        string, optional
        :param      task_completed:   Filter the completed tasks
        :type       task_completed:  boolean, optional

        :returns:   List of Task object with there edge
        :rtype:     List of dict {item: :class:`~aquarium.items.task.Task`, edge: :class:`~aquarium.edge.Edge`}
        """
        task = list()
        query = list()

        task.append("($Task")
        if task_completed:
            task.append("AND item.data.completion == 1")
        if task_status:
            task.append("AND item.data.status == '{0}'".format(task_status))
        if task_name:
            task.append("AND item.data.name == '{0}'".format(task_name))

        task.append(')')

        query.append("# <($Assigned)- ({0}".format(' '.join(task)))

        if project_key:
            query.append(
                "AND (<($Child, 5)- item._key == '{0}' AND path.vertices[*].type NONE == 'User')".format(project_key))

        query.append(")")

        result = self.traverse(meshql=' '.join(query))
        result = [self.parent.element(data) for data in result]
        return result



# -*- coding: utf-8 -*-
from . import JSON_CONTENT_TYPE
from .tools import evaluate, pretty_print_format
from dotmap import DotMap
import requests
import logging
logger=logging.getLogger(__name__)


class Entity(object):
    """
    This class describes an Entity object child of Aquarium class
    """
    def __init__(self, parent=None):
        """
        Constructs a new instance.

        :param      parent:  The parent
        :type       parent:  Aquarium instance
        """
        self.parent=parent
        self._key=None

    def __call__(self, data={}):
        """
        Callable to create new instance from a hierarchy or a key

        :param      data:  The data or the key
        :type       key:  string or int or dictionary

        :returns:   Entity instance
        :rtype:     Entity object
        """
        # Data variables

        inst=self.__class__(parent=self.parent)

        inst._key=''
        inst._id=''
        inst._rev=''
        inst.type=''
        inst.createdAt=''
        inst.updatedAt=''
        inst.createdBy=dict()
        inst.updatedBy=dict()
        inst.data=dict()

        if data:
            # As number or string
            if isinstance(data, str) or isinstance(data, int) or isinstance(data, type(u'')):
                inst._key=str(data)

            # As dict
            elif isinstance(data, dict):
                inst.set_data_variables(data=data)

        return inst

    def __str__(self):
        entity=vars(self).copy()
        entity.pop('parent', None)
        dash = '—' * ((len(self.__class__.__name__)) + 2)
        return '\n\t[%s]\n\t%s\n%s ' % (self.__class__.__name__, dash, pretty_print_format(entity, indent=8))
    def __repr__(self):
        return str(self)

    @property
    def session(self):
        return self.parent.session

    def set_data_variables(self, data={}):
        """
        Sets the data variables.

        :param      data:  The data
        :type       data:  dictionary
        """
        self._key=data.get('_key')
        self._id=data.get('_id')
        self._rev=data.get('_rev')
        self.type=data.get('type')
        self.createdAt=data.get('createdAt')
        self.updatedAt=data.get('updatedAt')
        self.createdBy=data.get('createdBy')
        self.updatedBy=data.get('updatedBy')

        entity_data=data.get('data')
        if entity_data:
            self.data=DotMap(entity_data)

    def do_request(self, *args, **kwargs):
        """
        Execute a request

        :param      args:    The arguments used to launch the process
        :type       args:    tuple
        :param      kwargs:  The keywords arguments used to launch the process
        :type       kwargs:  dictionary

        :returns:   request response
        :rtype:  list or dictionary
        """
        result=self.parent.do_request(*args, **kwargs)
        return result



# -*- coding: utf-8 -*-
import pprint
import logging
logger=logging.getLogger(__name__)
from .exceptions import RequestError, AuthentificationError,\
                        AutorisationError, PathNotFoundError, \
                        ConflictError, UploadExceedLimit, InternalError

def pretty_print_format(data={}, indent=8, width=80, depth=10):
    dict_string=pprint.pformat(data, indent=indent, width=width, depth=depth)
    return dict_string

def to_string_url(value=None):
    """
    Convert value to string url convention

    :param      value:  The value to convert
    :type       value:  number

    :returns:   The value converted
    :rtype:     string
    """
    return str(value).lower()

def evaluate(response=None):
    """
    Evaluate request response

    :param      response:        The response
    :type       response:        Response object

    :returns:   True if succeed
    :rtype:     boolean

    :raises     RuntimeError:  "error" value from response
    """
    status_code=response.status_code
    url=response.url
    logger.debug('Evaluate request response : status_code : %s / url : %s', status_code, url)
    if status_code == 200:
        return True
    elif status_code==400:
        raise RequestError(response)
    elif status_code==401:
        raise AuthentificationError(response)
    elif status_code==403:
        raise AutorisationError(response)
    elif status_code==404:
        raise PathNotFoundError(response)
    elif status_code==409:
        raise ConflictError(response)
    elif status_code==413:
        raise UploadExceedLimit(response)
    elif status_code==500:
        raise InternalError(response)
    else:
        raise RuntimeError('code {0} : {1} url:{2}'.format(
            status_code, response, url))




# -*- coding: utf-8 -*-
from ..item import Item


class Project(Item):
    """
    This class describes a project object child of Item class
    """

    def get_all(self, show_all=False):
        """
        Gets all projects accessible by the connected user

        :param      show_all:  Add completed and trashed projects
        :type       show_all:  boolean, optional

        :returns:   List of Project class
        :rtype:     List of :class:`~aquarium.items.project.Project`
        """
        query = list()
        query.append('# ($Project')

        if not show_all:
            query.append('AND item.data.completion != 1 AND NOT <($Trash)- *)')

        query.append(') SORT item.data.name ASC')
        result = self.parent.query(meshql=' '.join(query))
        result = [self.parent.cast(data) for data in result]
        return result

    def get_shots(self):
        """
        Gets all the shots of the project

        :returns:   List of Shot class, Edge class and Tasks
        :rtype:     List of dictionary {item: :class:`~aquarium.items.shot.Shot`, edge: :class:`~aquarium.edge.Edge`, tasks: [:class:`~aquarium.items.task.Task`]}
        """
        query = "# -($Child, 5)> 0, 0 $Shot VIEW $view"
        aliases = {
            "view": {
                "item": 'item',
                "edge": 'edge',
                "tasks": '# -($Child)> $Task VIEW item'
            }
        }
        result = self.traverse(meshql=query, aliases=aliases)
        result = [self.parent.element(data) for data in result]
        return result

    def get_assets(self):
        """
        Gets all the assets of the project

        :returns:   List of Asset class, Edge class and Tasks
        :rtype:     List of dictionary {item: :class:`~aquarium.items.asset.Asset`, edge: :class:`~aquarium.edge.Edge`, tasks: [:class:`~aquarium.items.task.Task`]}
        """
        query = "# -($Child, 5)> 0, 0 $Asset VIEW $view"
        aliases = {
            "view": {
                "item": 'item',
                "edge": 'edge',
                "tasks": '# -($Child)> $Task VIEW item'
            }
        }
        result = self.traverse(meshql=query, aliases=aliases)
        result = [self.parent.element(data) for data in result]
        return result



# -*- coding: utf-8 -*-
from . import JSON_CONTENT_TYPE
from .item import Item
from .edge import Edge
from .tools import evaluate
from .items.user import User
from .items.template import Template
from .items.project import Project
from .items.task import Task
from .items.shot import Shot
from .items.asset import Asset
from .items.usergroup import Usergroup
from .items.organisation import Organisation
from .items.playlist import Playlist
from .element import Element
from .utils import Utils

import requests

import sys
if sys.version_info[0] > 2:
    from urllib.parse import urljoin, urlparse
else:
    from urlparse import urljoin, urlparse

import json
import logging
logger=logging.getLogger(__name__)


class Aquarium(object):
    """
    This class describes the main class of Aquarium

    :param api_url: Specify the URL of the API.
    :type api_url: string
    :param token: Specify the authentication token, to avoid :func:`~aquarium.aquarium.Aquarium.signin`
    :type token: string, optional
    :param api_version: Specify the API version you want to use (default : `v1`).
    :type api_version: string, optional

    :var token: Get the current token (populated after a first :func:`~aquarium.aquarium.Aquarium.signin`)
    :var edge: Access to Edge class
    :vartype edge: :class:`~aquarium.edge.Edge`
    :var item: Access to Item class
    :vartype item: :class:`~aquarium.item.Item`
    :var asset: Access to Asset subclass
    :vartype asset: :class:`~aquarium.items.asset.Asset`
    :var playlist: Access to Playlist subclass
    :vartype playlist: :class:`~aquarium.items.playlist.Playlist`
    :var project: Access to Project subclass
    :vartype project: :class:`~aquarium.items.project.Project`
    :var shot: Access to Shot subclass
    :vartype shot: :class:`~aquarium.items.shot.Shot`
    :var task: Access to Task subclass
    :vartype task: :class:`~aquarium.items.task.Task`
    :var template: Access to Template subclass
    :vartype template: :class:`~aquarium.items.template.Template`
    :var user: Access to User subclass
    :vartype user: :class:`~aquarium.items.user.User`
    :var usergroup: Access to Usergroup subclass
    :vartype usergroup: :class:`~aquarium.items.usergroup.Usergroup`
    :var organisation: Access to Organisation subclass
    :vartype organisation: :class:`~aquarium.items.organisation.Organisation`
    :var utils: Access to Utils class
    :vartype utils: :class:`~aquarium.utils.Utils`
    """

    def __init__(self, api_url='', token='', api_version='v1'):
        """
        Constructs a new instance.
        """
        # Session
        self.session=requests.Session()

        self.api_url=api_url
        self.api_version=api_version
        self.token=token
        # Classes
        self.element=Element(parent=self)
        self.item=Item(parent=self)
        self.edge=Edge(parent=self)
        self.utils=Utils()
        # SubClasses
        self.user=User(parent=self)
        self.usergroup=Usergroup(parent=self)
        self.organisation=Organisation(parent=self)
        self.template=Template(parent=self)
        self.project=Project(parent=self)
        self.playlist=Playlist(parent=self)
        self.task=Task(parent=self)
        self.shot=Shot(parent=self)
        self.asset=Asset(parent=self)

    def do_request(self, *args, **kwargs):
        """
        Execute a request to the API

        :param      args:    Parameters used to send the request : HTTP verb, API endpoint
        :type       args:    tuple
        :param      kwargs:  Headers, data and parameters used for the request
        :type       kwargs:  dictionary

        :returns:   Request response
        :rtype:     List or dictionary
        """
        token=self.token

        decoding=True
        if 'decoding' in kwargs:
            decoding=kwargs.pop('decoding')

        if 'headers' in kwargs:
            headers=kwargs.pop('headers')
            if headers is not None:
                headers.update(dict(authorization=token))
        else:
            headers=dict(authorization=token)
            headers.update(JSON_CONTENT_TYPE)

        args=list(args)
        typ=args[0]
        path = self.api_url

        if len(args) > 1:
            is_files = args[1].find('/files/') >= 0
            if (is_files):
                path = urljoin(path, args[1])
            else:
                path = urljoin(path, '{api_version}/{endpoint}'.format(
                    api_version=self.api_version,
                    endpoint=args[1]
                ))
        else:
            path = urljoin(path, self.api_version)

        logger.debug('Send request : %s %s', typ, path)
        self.session.headers.update(headers)
        response=self.session.request(typ, path, **kwargs)
        evaluate(response)
        if decoding:
            response=response.json()
        return response

    def cast(self, data={}):
        """
        Creates an item or edge instance from a dictionary

        :param      data:         The object item or edge from Aquarium API
        :type       data:         dictionary

        :returns:   Instance of Edge or Item or items subclass
        :rtype:     :class:`~aquarium.edge.Edge` | :class:`~aquarium.item.Item` : [:class:`~aquarium.items.asset.Asset` | :class:`~aquarium.items.project.Project` | :class:`~aquarium.items.shot.Shot` | :class:`~aquarium.items.task.Task` | :class:`~aquarium.items.template.Template` | :class:`~aquarium.items.user.User` | :class:`~aquarium.items.usergroup.Usergroup`]
        """
        value=data
        #As Entity
        if data and '_id' in data.keys():
            id=data.get('_id')
            cls=None
            #As Item
            if id.split('/')[0]=='items':
                type=data.get('type')
                if type=='Project':
                    cls=self.project
                if type=='Playlist':
                    cls=self.playlist
                elif type=='User':
                    cls=self.user
                elif type=='Template':
                    cls=self.template
                elif type=='Usergroup':
                    cls=self.usergroup
                elif type=='Asset':
                    cls=self.asset
                elif type=='Shot':
                    cls=self.shot
                elif type=='Task':
                    cls=self.task
                elif type=='Organisation':
                    cls=self.organisation
                else:
                    cls=self.item
            #As Edge
            elif id.split('/')[0]=='connections':
                cls=self.edge
            if cls is not None:
                value=cls(data=data)

        return value

    def signin(self, email='', password=''):
        """
        Sign in a user with its email and password

        :param      email:     The email of the user
        :type       email:     string
        :param      password:  The password of the user
        :type       password:  string
        """
        return self.user.signin(email=email, password=password)

    def connect(self, email='', password=''):
        """
        Alias of :func:`~aquarium.aquarium.Aquarium.signin`
        """
        return self.user.signin(email=email, password=password)

    def signout(self):
        """
        Sign out current user by clearing the stored authentication token

        .. note::
            After a :func:`~aquarium.aquarium.Aquarium.signout`, you need to use a :func:`~aquarium.aquarium.Aquarium.signin` before sending authenticated requests

        :returns: None
        """
        logger.info('Disconnect current user')
        logger.debug('Clear authentication token for logout')
        self.user.signout()

    def logout(self):
        """
        Alias of :func:`~aquarium.aquarium.Aquarium.signout`
        """
        self.signout()

    def me(self):
        """
        Alias of :func:`~aquarium.aquarium.Aquarium.get_current_user`


        :returns:   A :class:`~aquarium.items.user.User` instance of the connected user.
        :rtype:     :class:`~aquarium.items.user.User` object
        """
        return self.get_current_user()

    def get_current_user(self):
        """
        Alias of :func:`~aquarium.items.user.User.get_current`


        :returns:   A :class:`~aquarium.items.user.User` instance of the connected user.
        :rtype:     :class:`~aquarium.items.user.User` object
        """
        result=self.user.get_current()
        return result

    def mine(self):
        """
        Alias of :func:`~aquarium.items.user.User.get_profile`


        :returns:   User, Usergroups and Organisations object
        :rtype:     Dict {user: :class:`~aquarium.items.user.User`, usergroups: [:class:`~aquarium.items.usergroup.Usergroup`], organisations: [:class:`~aquarium.items.organisation.Organisation`]}
        """
        return self.user.get_profile()

    def get_server_status(self):
        """
        Gets the server status.

        :returns:   The server status
        :rtype:     dictionary
        """
        result=self.do_request('GET', 'status')
        return result

    def upload_file(self, path=''):
        """
        Uploads a file on the server

        .. note::
            The file is just uploaded to Aquarium. The metadata are not saved on any item. Use :func:`~aquarium.item.Item.update_data` to save them on an item.
            You can also directly upload a file on an item with :func:`~aquarium.item.Item.upload_file`.

        :param      path:  The path of the file to upload
        :type       path:  string

        :returns:   The file metadata on Aquarium
        :rtype:     dictionary
        """
        logger.debug('Upload file : %s', path)
        files=dict(file=open(path, 'rb'))
        result = self.do_request('POST', 'upload', headers={'Content-Type': None}, files=files)
        files['file'].close()
        return result

    def query(self, meshql='', aliases={}):
        """
        Query entities

        .. tip::
            For better performances, we advice you to use the function :func:`~aquarium.item.Item.traverse`

        :param      meshql:        The meshql string
        :type       meshql:        string
        :param      aliases:       The aliases used in the meshql query
        :type       aliases:       dictionary

        :returns:   List of item, edge or VIEW used in the meshql query
        :rtype:     list
        """
        logger.debug('Send query : meshql : %s / aliases : %r',
                     meshql, aliases)
        data=dict(query=meshql, aliases=aliases)
        result=self.do_request('POST', 'query', data=json.dumps(data))
        return result

    def get_file(self, file_path):
        """
        Get stored file on Aquarium server

        :param      file_path:     The file path from item property (Exemple: `/files/file_id.jpg`)
        :type       file_path:     string

        :returns:   The file
        :rtype:     list
        """

        response = self.do_request('GET', file_path, decoding=False)
        return response.content



# -*- coding: utf-8 -*-
import json
from .. import URL_CONTENT_TYPE
from ..item import Item
import logging
logger = logging.getLogger(__name__)


class Usergroup(Item):
    """
    This class describes an user group object child of Item class.
    """

    def get_users(self):
        """
        Get the users of the user group

        :returns:   List of User object
        :rtype:     List of :class:`~aquarium.items.user.User`
        """
        result = self.do_request(
            'GET', 'usergroups/'+self._key, headers=URL_CONTENT_TYPE)

        result = [self.parent.cast(data) for data in result]
        return result

    def add_user(self, user_key=''):
        """
        Add an user in the user group

        :param      user_key:  The user key
        :type       user_key:  string

        :returns:   Membership edge object
        :rtype:     dictionary
        """
        logger.debug('Add user %s to usergroup %s', user_key, self._key)
        payload = dict(userKey=user_key)
        result = self.do_request(
            'POST', 'usergroups/'+self._key, data=json.dumps(payload))
        return result

    def remove_user(self, user_key=''):
        """
        Remove an user from the user group

        :param      user_key:  The user key
        :type       user_key:  string

        :returns:   Deleted membership edge object
        :rtype:     dictionary
        """
        logger.debug('Remove user %s to usergroup %s', user_key, self._key)
        payload = dict(userKey=user_key)
        result = self.do_request(
            'DELETE', 'usergroups/'+self._key, data=json.dumps(payload))
        return result



# -*- coding: utf-8 -*-
from .. import URL_CONTENT_TYPE
from ..item import Item
import logging
logger = logging.getLogger(__name__)


class Template(Item):
    """
    This class describes a Template object child of Item class.
    """

    def apply(self):
        """
        Apply the template on all its items.

        :returns:   Item template
        :rtype:     :class:`~aquarium.items.template.Template`
        """
        logger.debug('Apply template %s', self._key)
        result = self.do_request(
            'POST', 'templates/'+str(self._key)+'/apply', headers=URL_CONTENT_TYPE)
        result = self.parent.cast(result)
        return result



# -*- coding: utf-8 -*-
import os
from ..item import Item


class Asset(Item):
    """
    This class describes an Asset object child of Item class.
    """

    def upload_on_task(self, task_name='', path='', data={}, version_name=None, override_media = True, message = None):
        """
        Uploads new media version on asset task

        :param      task_name:      The task name
        :type       task_name:      string
        :param      path:           The media path to upload
        :type       path:           string
        :param      data:           The data you want to upload with the file, optional
        :type       data:           dict
        :param      version_name:   The name of the version where you want to upload the file. Without version_name, the media is uploaded in the task, optional
        :type       version_name:   string
        :param      override_media: If the media already exist, a new history is created to override the existing one, optional
        :type       override_media: boolean
        :param      message:        The message associated with the upload, optional
        :type       message:        string

        :returns:   Updated media object
        :rtype:     dictionary
        """
        file_dir, file_name = os.path.split(path)
        query = "# -($Child, 2)> 0,1 $Task AND item.data.name == '{0}' VIEW $view".format(
            task_name)

        aliases={
            'view':{
                "item": "item",
                "mediaKey": "FIRST(# -($Child)> 0,1 $Media VIEW item._key)"
            }
        }

        tasks=self.traverse(meshql=query, aliases=aliases)
        if not tasks or len(tasks) == 0:
            raise RuntimeError('Could not find request')

        media_key = tasks[0].get('mediaKey')
        task = self.parent.cast(tasks[0]['item'])

        if version_name == None:
            if not media_key or override_media == False:
                return task.append(type='Media', data=data, path=path)
            else:
                return self.parent.item(media_key).upload_file(path=path, data=data, message=message)
        else:
            versions = task.get_children(types='Version', names=version_name)

            if len(versions) > 0:
                version = versions[0].item
            else:
                version = task.append(
                    type='Version', data=dict(name=version_name)).item

            if override_media:
                medias = version.get_children(types='Media')

                if len(medias) > 0:
                    existing_medias = [media for media in medias if media.item.data.originalname == file_name]
                    if len(existing_medias) > 0:
                        media = existing_medias[0].item
                        return media.upload_file(
                            path=path, data=data, message=message)
                    else:
                        return version.append(type='Media', data=data, path=path)
                else:
                    return version.append(type='Media', data=data, path=path)
            else:
                return version.append(type='Media', data=data, path=path)


    def get_tasks(self, task_name='', task_status=''):
        """
        Gets the tasks of the asset

        :param      task_name:    The name of the task
        :type       task_name:    string, optional
        :param      task_status:  The status of the task
        :type       task_status:  string, optional

        :returns:   List of Task object and Edge object
        :rtype:     List of dictionary {item: :class:`~aquarium.items.task.Task`, edge: :class:`~aquarium.edge.Edge`}
        """
        query='# -($Child)> $Task'
        if task_name:
            query+=" AND item.data.name == '{0}'".format(task_name)
        if task_status:
            query+=" AND item.data.status == '{0}'".format(task_status)

        result=self.traverse(meshql=query)
        result=[self.parent.element(data) for data in result]
        return result

    def get_assigned_tasks(self, user_key= '', task_name='', task_status=''):
        """
        Gets the asset's assigned tasks to specific user

        :param      user_key:     The user key
        :type       user_key:     string
        :param      task_name:    The name of the task used to filter
        :type       task_name:    string, optional
        :param      task_status:  The status of the task used to filter
        :type       task_status:  string, optional

        :returns:   List of Task object and Edge object
        :rtype:     List of dictionary {item: :class:`~aquarium.items.task.Task`, edge: :class:`~aquarium.edge.Edge`}
        """
        query="# -($Child)> $Task"

        if task_name:
            query+=" AND item.data.name == '{0}'".format(task_name)
        if task_status:
            query+=" AND item.data.status == '{0}'".format(task_status)

        query+=" AND -($Assigned)> item._key == '{0}'".format(user_key)

        result=self.traverse(meshql=query)
        result=[self.parent.element(data) for data in result]
        return result

    def get_by_task(self, project_key='', task_status='', task_name='', task_completed=False):

        """
        Gets project tasks by filters.

        :param      project_key:     The project key
        :type       project_key:     string
        :param      task_status:     The task status
        :type       task_status:     string
        :param      task_name:       The task name
        :type       task_name:       string, optional
        :param      task_completed:  Task is completed
        :type       task_completed:  boolean

        :returns:   The tasks.
        :rtype:     dictionary
        """
        project=list()
        task=list()

        project.append("(<($Child, 5)- item._key == '{0}' AND path.vertices[*].type NONE == 'User')".format(project_key))

        task.append("(-($Child)> ($Task AND item.data.status == '{0}'".format(task_status))
        task.append("AND item.data.completion {0}= 1".format(['!', '='][task_completed]))

        if task_name:
            task.append("AND item.data.name == '{0}'".format(task_name))
        task.append('))')

        query="# $Asset AND {0} AND {1}".format(' '.join(project), ' '.join(task))

        result=self.parent.query(meshql=query)
        result=[self.parent.element(data) for data in result]
        return result



# -*- coding: utf-8 -*-
from ..item import Item
import logging
logger = logging.getLogger(__name__)


class Playlist(Item):
    """
    This class describes a Playlist object child of Item class.
    """

    def get_medias(self, track=None):
        """
        Gets the medias of the playlist

        :param      track:    The ID of the track. If no track specified, all medias will be returned
        :type       track:    integer (0 or 1), optional (default: None)

        :returns:   List of Media object
        :rtype:     List of dictionary {media: :class:`~aquarium.items.media.Media`, track: integer, versionKey: integer}
        """

        filter = ''
        if track != None:
            filter = 'AND edge.data.track == {0}'.format(track)

        query = '# -($Child OR $Playlist)> 0,5000 $Media {0} UNIQUE SORT edge.createdAt ASC VIEW $view'.format(
            filter)
        aliases = {
            'view': {
                'media': 'item',
                'track': 'edge.data.track',
                'versionKey': 'edge.data.versionKey'
            }
        }

        result=self.traverse(meshql=query, aliases=aliases)
        result=[self.parent.element(data) for data in result]
        return result

    def get_playlists(self):
        """
        Gets the playlists added in this playlist

        :returns:   List of Playlist object
        :rtype:     List of :class:`~aquarium.items.playlist.Playlist`
        """

        query = '# -($Child)> $Playlist VIEW item'

        result=self.traverse(meshql=query)
        result = [self.parent.cast(item) for item in result]
        return result

    def import_medias(self, media_paths, track=0):
        """
        Import media files into the playlist

        :param      media_paths:    List of media file paths that need to be imported in the playlist
        :type       media_paths:    list, optional
        :param      track:          The ID of the track where the media will be imported. If no track specified, media are imported in the first track
        :type       track:          integer (0 or 1), optional (default: 0)

        :returns:   List of Media object
        :rtype:     List of :class:`~aquarium.items.media.Media`
        """
        medias = []
        if track > 1:
            track = 1
        elif track < 0:
            track = 0

        for file in media_paths:
            edge_data = {
                'track': track
            }
            media = self.append(type='Media', path=file, edge_data=edge_data)
            medias.append(media.item)

        return medias

    def add_playlist(self, playlist_key):
        """
        Add an existing playlist into the current playlist

        :param      playlist_key: The playlist _key to add in the current playlist
        :type       playlist_key: string

        :returns:   The created edge between the current playlist and the imported one
        :rtype:     :class:`~aquarium.edge.Edge`
        """

        return self.parent.edge.create(type='Child', from_key=self._key, to_key=str(playlist_key))

    def remove_media(self, media_key):
        """
        Remove media from the playlist

        :param      media_key:    Item media _key to remove from the playlist
        :type       media_key:    string

        :returns:   None
        :rtype:     None
        """

        query = '# -($Child OR $Playlist)> 0,1 $Media AND item._key == "{0}" VIEW edge'.format(media_key)
        media_edge = self.traverse(meshql=query)

        if len(media_edge) > 0:
            edge = self.parent.cast(media_edge[0])
            edge.delete()

    def remove_playlist(self, playlist_key):
        """
        Remove the imported playlist from the current playlist

        :param      media_key:    Item playlist _key to remove from the playlist
        :type       media_key:    string

        :returns:   None
        :rtype:     None
        """

        query = '# -($Child OR $Playlist)> 0,1 $Playlist AND item._key == "{0}" VIEW edge'.format(playlist_key)
        playlist_edge = self.traverse(meshql=query)

        if len(playlist_edge) > 0:
            edge = self.parent.cast(playlist_edge[0])
            edge.delete()

    def change_media_track(self, media_key, track=0):
        """
        Move the media to another track

        :param      media_key:    Item media _key to remove from the playlist
        :type       media_key:    string
        :param      track:        The ID of the track where the media will be imported. If no track specified, media are imported in the first track
        :type       track:        integer (0 or 1), optional (default: 0)

        :returns:   None
        :rtype:     None
        """

        query = '# -($Child OR $Playlist)> 0,1 $Media AND item._key == "{0}" VIEW edge'.format(media_key)
        media_edge = self.traverse(meshql=query)

        if len(media_edge) > 0:
            edge = self.parent.cast(media_edge[0])
            data = {
                'track': track
            }
            edge.update_data(data=data)




# -*- coding: utf-8 -*-
from __future__ import absolute_import
import logging
from logging import NullHandler


JSON_CONTENT_TYPE={"Content-Type": "application/json"}
URL_CONTENT_TYPE={"Content-Type": "application/x-www-form-urlencoded"}

DEFAULT_STATUSES=dict(
	toDo={'status':'TO DO', 'color': '#FBEB06', 'valid': False, 'completion': 0},
	wip={'status':'WIP', 'color': '#F3A215', 'valid': False, 'completion': 0.3},
	rtk={'status':'RTK', 'color': '#f35415', 'valid': False, 'completion': 0.5},
	pendingReview={'status':'PENDING REVIEW', 'color': '#15c8f3', 'valid': False, 'completion': 0.9},
	done={'status':'DONE', 'color': '#9cde4d', 'valid': True, 'completion': 1},
	cancelled={'status':'CANCELLED', 'color': '#918F89', 'valid': False, 'completion': 0}
)

# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(NullHandler())

from .aquarium import Aquarium
